#include <PSX/System/IOPorts/interrupt_io.hpp>
#include <PSX/System/IOPorts/timer_io.hpp>
#include <PSX/System/syscalls.h>

#include <GPU/gpu.hpp>
#include <stdio.h>

namespace JabyEngine {
    namespace Timer {
        extern InterrupCallback IRQCallback;

        static constexpr double CPU_Frequency_Hz      = 33868800.0;
        static constexpr double CPU_Frequncey_Hz_Div8 = (CPU_Frequency_Hz/8.0);

        template<typename T>
        static constexpr T NS_Per_Tick(double CPU_Frequency_Hz) {
            return static_cast<T>((1.0/CPU_Frequency_Hz)*1000.0*1000.0*1000.0);
        }

        template<typename T>
        static constexpr T US_Per_Tick(double CPU_Frequency_Hz) {
            return static_cast<T>((1000.0/NS_Per_Tick<double>(CPU_Frequency_Hz)));
        }

        template<typename T>
        static constexpr T MS_Per_Tick(double CPU_Frequency_Hz) {
            return static_cast<T>(((1000.0*1000.0)/NS_Per_Tick<double>(CPU_Frequency_Hz)));
        }

        void setup() {
            using namespace Timer_IO;

            static constexpr auto     Mode   = CounterMode::with(CounterMode::FreeRun, Counter2::SyncMode::Freerun, CounterMode::ResetAfterTarget, CounterMode::IRQAtTarget, CounterMode::IRQEveryTime, CounterMode::IRQPulse, Counter2::Source::System_Clock_Div_8);
            static constexpr uint16_t Target = MS_Per_Tick<uint16_t>(CPU_Frequncey_Hz_Div8)*10;

            Interrupt::disable_irq(Interrupt::Timer2);

            __syscall_EnterCriticalSection();
                __syscall_SysEnqIntRP(Timer2Irq, &IRQCallback);
            __syscall_ExitCriticalSection();

            Counter2.target.write(CounterTarget::CounterTargetValue.with(Target));
            Counter2.mode.write(Mode);

            Interrupt::enable_irq(Interrupt::Timer2);
        }
    }
}